from configobj import ConfigObj
from validate import Validator
import logging

import LegoTGEntities

class LegoTGConfigHandler:
  # Configuration must follow this spec:
  configspec = """
    [logging]
	log_file = string(max=80, default=None)
	log_level = integer(0,2, default=2)

    [nodes]
      user = string(max=255, default=None)
      gateway = string(max=255, default=None)
      key_filename = force_list()
      reject_unknown_hosts = boolean(default=False)
      shell = string(max=255, default=None)
      ssh_config_path = string(max=255, default=None)
      timeout = integer(default=10)
      hosts = force_list()
      
      [[__many__]]
        user = string(max=255, default=None)
        gateway = string(max=255, default=None)
        key_filename = force_list()
        reject_unknown_hosts = boolean(default=False)
        shell = string(max=255, default=None)
        ssh_config_path = string(max=255, default=None)
        timeout = integer(default=10)
        hosts = force_list()

    [groups]

    [extraction]
        order = force_list()
        install_order = force_list()
        setup_order = force_list()
        run_order = force_list()
        cleanup_order = force_list()

	[[__many__]]
	def = string(max=80)
	target = force_list(default=None)
	[[[args]]]
        
    [resource allocation]
        order = force_list()
        install_order = force_list()
        setup_order = force_list()
        run_order = force_list()
        cleanup_order = force_list()

	[[__many__]]
	def = string(max=80)
	target = force_list(default=None)
	[[[args]]]

    [experiment]
        order = force_list()
        install_order = force_list()
        setup_order = force_list()
        run_order = force_list()
        cleanup_order = force_list()

	[[__many__]]
	def = string(max=80)
	target = force_list(default=None)
	[[[args]]]

  """.split('\n')

  logger = None
  config = None
  config_file = None
  entities = None

  def __init__(self, config_file):
    # Check we can open the config file.
    try:
      with open(config_file) as f: pass
    except IOError as e:
      raise e
    try:
      self.config = ConfigObj(config_file, configspec=self.configspec, file_error=False, interpolation='Template')
    except Exception as e:
      raise e
    val = Validator()
    test = self.config.validate(val)
    if test == False:
      raise RuntimeError('Config file contains errors') 
    self.config_file = config_file  
    try:
      self.logger = self.set_logger()
    except Exception as e:
      raise e
    try:   
      self.entities = LegoTGEntities.LegoTGEntities(self.logger, self.config['nodes']['physical'], self.config['nodes']['virtual'], self.config['nodes']['topoxml'], self.config['groups'])
    except Exception as e:
      raise e
      
  def set_logger(self):
    # Set up a logger.
    file = self.config['logging']['log_file']
    if file != None:
      try:
        lvl = int(self.config['logging']['log_level'])
        logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p', filename=self.config['logging']['log_file'], level=lvl, filemode='a+')
      except Exception as e:
        raise e
    self.logger = logging.getLogger('Orchestrator')
    return self.logger
  
  def get_def_file(self, name, section):
    if section in ['experiment']:
      return self.config['experiment'][name]['def']
    return []
  
  def get_var_value_pairs(self, name, section):
    if section in ['experiment', 'extraction', 'allocation']:
      return self.config['experiment'][name]['args']
    return []

  def get_block_list(self, section):
    block_list = []
    if section in ['experiment', 'extraction', 'allocation']:
      for object in self.config[section]:
        # We may have variables (such as order) or a section
        # We're looking for the section names
        if is_section(config[section][object]):
          block_list.append 

      if not isinstance(self.config[section]['order'], list):
        single = self.config[section]['order']
        self.config[section]['order'] = list()
        self.config[section]['order'].append(single)
      return self.config[section]['order']
    self.logger.warn("Do not know about section \"%s\". No block list given." % (section))
    return block_list
    
  def get_host_target_list(self, name, section):
    if section in ['experiment']:
      if not isinstance(self.config[section][name]['target'], list):
        single = self.config[section][name]['target']
        self.config[section][name]['target'] = set()
        self.config[section][name]['target'].add(single)
      for target in self.config[section][name]['target']:
        # Target may be a group, in which case we need the group members.
        if len(self.entities.return_members(target)) != 0:
          return self.entities.return_members(target)
        # Or target may just be a host.
        else:
          return [target]  
    self.logger.warn("No targets for block \"%s\" in Section \"%s\"" % (name, section))
        
  def block_section_exists(self, name, section):
    if name in self.config[section]:
      return True
    else:
      return False
          
  def print_groups(self):
    for entity in self.config['groups']:
      print entity
      for e in self.config['groups'][entity]:
        print "\t%s" % e
      print
