from fabric.state import commands, connections
import random
import threading
import re
import time
import logging
import signal
import sys, time
import argparse
import OrchestratorHostsHandler
import OrchestratorGroupHandler
import OrchestratorBlockHandler
import OrchestratorConfgHandler

def splitbynum(x):
  r = re.compile('(\d+)')
  l = r.split(x)
  return[int(y) if y.isdigit() else y for y in l]


class Orchestrator:
  def __init__(self, args):
    self.args = args
    # Get our experiment configuration parser set up.
    try: 
      self.config_handler = OrchestratorConfgHandler.OrchestratorConfgHandler(args.config_file, args.log_lvl)
    except:
      raise
    self.logger = logging.getLogger('Orchestrator')
    logger = logging.getLogger('Orchestrator')
    logger.info('')
    logger.info('********************** Beginning Experiment **********************')
    
    # Set up our host handler.
    self.host_handler = OrchestratorHostsHandler.OrchestratorHostsHandler(self.config_handler.return_hosts_config())
    # Test and collect passwords for hosts.
    if args.checks:
      print("\n\033[1mTesting connections and collecting passwords for hosts.\033[0m")
      if not self.host_handler.check_hosts():
        raise Exception('Could not log into all hosts.')
    if args.tmux_install == True:
      self.host_handler.install_tmux()
    if not self.host_handler.setup_detach_handler(self.config_handler.return_exp_name()):
      raise Exception('Could not set up detach handler.')
  
    # Set up our group handler.
    try:
      print("\n\033[1mParsing user defined groups.\033[0m")
      self.group_handler = OrchestratorGroupHandler.OrchestratorGroupHandler(self.config_handler.return_group_config())
    except:
      raise
    # Get a software block handler.
    if args.phase == 'extract':
      section = 'extraction'
    elif args.phase =='allocate':
      section ='resource allocation'
    elif args.phase == 'run':
      section ='experiment'
    else:
      raise Exception('Unknown experiment phase. Exiting...')
      exit()
    print("\033[1mFinding and loading block software for phase %s.\033[0m" % (section))
    self.block_handler = OrchestratorBlockHandler.OrchestratorBlockHandler(self.config_handler.return_block_config(section), section)
    try:
      exp_dir = self.config_handler.return_exp_dir()
      exp_name = self.config_handler.return_exp_name()
      if exp_dir != None and exp_dir != '':
        self.block_handler.load_all_blocks(exp_dir=exp_dir, exp_name=exp_name)
      else:
        self.block_handler.load_all_blocks()
    except:
      raise
  
  def get_experiment_name(self):
    return self.config_handler.return_exp_name() 
  
  def list_all_nodes(self):
    return self.host_handler.list_hosts()
  
  def get_action_list(self):        
    # Were we given an action/actions on the command line?
    if self.args.action != None:
      self.logger.info('Given action items on command line (ignoring any actions in config): %s' % (self.args.action))    
      action_list = self.args.action
    # Else, do we have an action list in the config?
    else:
      action_list = self.block_handler.action_order()
    return action_list
  
  def get_block_list(self):
    try:
      blocks = self.block_handler.block_call_order()
    except Exception as e:
      raise e
    return blocks

  def get_target_hosts(self, block):
    """
      Gives the hosts a block should be run on.
      Returns None, [] or list of full host names.
      Returns an empty list if no tarets are specified for the block.
      It may be desirable to skip use of a block by commenting out targets.
    """
    # Get host targets for this block
    # Get the target (could be hosts or groups or both):
    targets = self.block_handler.get_target(block)
    target_hosts = []
    
    # Sort out the groups.
    for target in targets:
      if target == '':
        continue
      # Is the target a group? Or a host?
      if self.group_handler.is_group(target):    
        # GROUP:
        # Get the full host name for all members of the group.
        for host in self.group_handler.host_members(target):
          # For each host, get the full name (and check that it's really a host we know about.)
          full_name = self.host_handler.full_name(host)
          if full_name != None:
            target_hosts.append(full_name)
          else:
            self.logger.error('Could not parse targets for block \'%s\'. Unknown target \'%s\'' % (block, host))
            raise Exception('Could not parse targets.')
      elif self.host_handler.is_host(self.host_handler.full_name(target)):  
        # HOST: This target wasn't a group, so we assume it's a host.
        full_name = self.host_handler.full_name(target)
        if target != None:
          target_hosts.append(full_name)
        else:
          self.logger.error('Could not parse targets for block \'%s\'. Unknown target \'%s\'' % (block, target))          
          raise Exception('Could not parse targets.')
      else:
        self.logger.error('Unknown target \'%s\'' % (target))
        raise Exception('Could not parse targets.')

    # We may have some duplicates after all the group expanding, so remove dups.
    target_hosts = list(set(target_hosts))
    target_hosts.sort(key=splitbynum)
    self.logger.info ("Block: %s -> Targets: %s"% (block, target_hosts))
    return target_hosts
  
  def get_block_target_dict(self):
    bt_dict = {}
    blist = self.get_block_list()
    for b in blist:
      tlist = self.get_target_hosts(b)
      bt_dict[b] = tlist
    return bt_dict
  
  def do_action(self, block, action):
    # Get host targets for this block
    try:
      target_hosts = self.get_target_hosts(block)
    except:
      raise
    
    if target_hosts == None or target_hosts == []:
      self.logger.warn("Action \'%s\', Block \'%s\': No targets. Nothing to do." % (action, block))
      return
    
    self.logger.info("Action \'%s\', Block \'%s\', Targets: \'%s\'"% (action, block, target_hosts))
    # Set the GROUP variable for this run, to let hosts know who else is doing the same block.
    try:
      self.block_handler.set_group_attr(block, target_hosts)
    except:
      raise
    try:
      # Set the BLOCK_TARGET attribute (dictionary), if the block asks for it.
      if self.block_handler.asks_for_block_target_attr(block):
        self.logger.info("Block \'%s\' asks for BLOCK_TARGET dictionary to be set." % (block))
        block_target_dict = self.get_block_target_dict()
        self.block_handler.set_block_target_attr(block, block_target_dict)
    except:
      raise
    # Get function for this action
    try:
      function = self.block_handler.action_function(block, action)
    except:
      raise

    # Copy input files if we need to.
    #if action == 'install' or action == 'setup':
    #  var_value_inputs = self.block_handler.get_var_value_pairs(block, type='input')
    #  i = 0
    #  for host in target_hosts:
    #    interpolation_dict = {}
    #    interpolation_dict['host'] = host.split('.')[0]
   #     interpolation_dict['part'] = str(i)
   #     self.host_handler.copy_files(host, var_value_inputs, interpolation_dict, type='input')
   #     i = i + 1
    

    # Run action/function on targets!
    try:
      if function != None:
        self.host_handler.run_on_host(target_hosts, function, self.config_handler.return_exp_name(), block, action)
    except:
      raise
    
  def wait_for_action(self, block, action):
    """
    When using detach services, this program (orchestrator) does not block while the action
    completes on the experiment nodes.
  
    Instead, the actions are left to complete in the tmux session.
  
    Here we check on progress.
    """
    if block == None and action == None:
      # We're not waiting on anything. Return!
      return
    
    if block == None and action != None:
      # We're waiting on everything associated with an action, so we ignore block.
      # Get list of block-actions we should kill in the event they're the only ones left running.
      # eg. tcpdump should be killed if all else is done.
      exclusion_list = self.block_handler.blocks_to_force_kill(action)
      print("Looking for tasks to wait on")
      while(self.host_handler.action_not_done(action, exclusion_list, self.config_handler.return_exp_name())):
        print ("\033[1mWaiting on all blocks to finish %s.\033[0m" % (action))
        time.sleep(5)
      # Kill left over blocks on the force kill list:
      if exclusion_list != []:
        print("Other blocks are done. Now stopping blocks which need force kill.")
        self.logger.info("Other blocks are done. Now stopping blocks which need force kill.")
        for b in exclusion_list:
          print("Stopping %s" % b)
          self.logger.info("Stopping %s" % b)
          self.do_action(b, 'stop')
    elif block != None and action != None:
      while(self.host_handler.action_not_done(action, [], self.config_handler.return_exp_name(), block=block)):
        print("Waiting on tasks related to %s." % (block))
        time.sleep(5)
  
  def clean_up(self):
    """
    Clean up any state and files on experiment nodes, and locally.
    """
    try:
      self.host_handler.clean_up(self.get_experiment_name())
    except Exception as e:
      raise(e)

sigint_flag = False
  
def main():
  global sigint_count
  sigint_count = 0
  global sigint_flag
  original_sigint = signal.getsignal(signal.SIGINT)
  
  def ignore(signum, frame):
    signal.signal(signal.SIGINT, original_sigint)
    print "Are you suuuuure????"
    return
  
  def signal_handler(signum, frame):
    signal.signal(signal.SIGINT, ignore)
    global sigint_flag
    sigint_flag = True
    done(abnormal=True)
  
  signal.signal(signal.SIGINT, signal_handler)
  
  try:
    block_list = orchestrator.get_block_list()
  except Exception as e:
    print(e)
    done(abnormal=True)

  if block_list != None:
    logger.info("List of blocks to call is: %s " % (block_list))
    try:
      action_list = orchestrator.get_action_list()    
    except Exception as e:
      print(e)
      done(abnormal=True)
    if action_list != None:
      action_index = 0
      for action in action_list:
        print("\033[1m%s:\033[0m" % (action))
        # We need to wait till all blocks running this action are done
        # before we move on.
        previous_action = get_previous_action(action_list, action_index)
        print("Waiting on action %s" % (previous_action))
        orchestrator.wait_for_action(None, previous_action)
        for block in block_list:
          print("\033[1m\t*******************[ %s (%s) ]******************\033[0m" %(block, action))
          # Does this block depend on other blocks?
          # If so, we have to wait for these blocks to finish before starting this one.
          # XXX Should add sanity checking that block order has depends called before
          # dependent block.
          depends_list = orchestrator.block_handler.get_depends_list(block, action)
          if depends_list != []:
            print("Block %s depends on blocks:" % (block))
          for d in depends_list:
            print("\t%s. Waiting on %s" %(d, action))
            orchestrator.wait_for_action(d, action)
          try:
            orchestrator.do_action(block, action)  
          except Exception as e:
            print(e)
            done(abnormal=True)
          if sigint_flag:
            print("Quitting")
            break
        action_index = action_index + 1
        last_action = get_previous_action(action_list, action_index)
        orchestrator.wait_for_action(None, last_action)
        if sigint_flag:
          print("Quitting")
          break
  else:
    logger.warn("No software blocks specified. Nothing to do.")
    print("Nothing to do.")
  print("\033[1m\tDone running actions.\033[0m")
  if sigint_flag:
    done(abnormal=True)
  done()
  
def get_previous_action(action_list, action_index):
  print action_list
  if action_index == 0:
    return None
  return action_list[action_index-1]

def done_from_signal():
  done_lock = threading.RLock()
  with done_lock:
    done(abnormal=True)

def done(abnormal=False):
  # Really hacky way for us to run done only once
  # We seed random in main with a fixed number, and therefore know the 
  # sequence of "random" numbers that will be generated. The first will be
  # 15...15 won't show up in the sequence again for quite some time.
  r = random.randrange(0,25)
  if r != 15:
    return
  print("\033[1mCleaning up.\033[0m")
  logger = logging.getLogger('Orchestrator')
  
  # Allow orchestrator to clean things up:
  try:
    orchestrator.clean_up()
  except Exception as e:
    print(e)
    pass
  if abnormal == True:
    # We exited due to an error.
    logger.warn("Aborting...")
  print("Cleaning up Fabric state.")
  for key in connections.keys():
    logger.info("Disconnecting from %s" %(key))
    connections[key].close()
    del connections[key]
  print("Cleaning up logger.")
  logger.info("********************** End of Experiment **********************")
  logging.shutdown()
  exit()

def phase_type(string):
  if string not in ['extract', 'allocate', 'run']:
    raise TypeError
  return string


# Global orchestrator:
parser = argparse.ArgumentParser(description='Software orchestrator to coordinate LegoTG')
parser.add_argument("-p", "--phase", dest='phase', required=True, type=str, default='run', choices=['extract', 'allocate', 'run'])
parser.add_argument("-b", "--blocks", dest='blocks', nargs='+', help="Block or list of blocks to perform actions on. Overrides blocks listed in the 'order' var in experiment config.")
parser.add_argument("-a", "--action", dest='action', nargs="+", help='Action or list of actions to perform. Overrides default action list in experiment config.')
parser.add_argument("-l", "--level" , dest='log_lvl', help='Set log level. Overrides log level set in experiment configuration file.')
parser.add_argument("-c", "--skipcheck" , dest='checks', action="store_false")
parser.add_argument("-f", "--config", dest='config_file', type=str, required=True, help="LegoTG experiment configuration file.")
parser.add_argument("-t", "--tmux", dest='tmux_install', action='store_false', default=True, help="Install tmux on hosts with env var tmux set using apt-get")
args = parser.parse_args()

try:
  orchestrator = Orchestrator(args)
  logger = logging.getLogger('Orchestrator')
except Exception as e:
  print('Problem setting up Orchestrator.')
  print(e)
  print('Aborting...')
  exit()

if __name__ == "__main__":
  random.seed(5)
  main()
  
  
  