import inspect, imp, os, sys, warnings, logging
import configobj
from fabric.api import hosts, run, task, env, settings
from fabric.tasks import execute


import OrchestratorGlobals


class OrchestratorBlockHandler:
  """
  Given a config for a specific section ("[extraction]", "[resource allocation]" and "[experiment]"),
  handles config parsing for this sections and loading block interface code.
  """
  logger = None
  blocks = {}
  config = None
  section = ''
    
  def __init__(self, config, section='Not Specified'):
    """
      Creates an object which handles blocks in a given experiment section.
      
      Arguments: 
      config -- the configuration section to work from ("[extraction]", "[resource allocation]" and "[experiment]")
      section -- the name (string) of the section for logging purposes. Does not need to be specified.
    """
    self.section = section
    self.config = config
    self.logger = logging.getLogger('Orchestrator')
  
  def block_def_check(self, block):
    """Sanity checks and returns the path to the block definition."""
    try:
      blk_def = self.config[block]['def']
    except Exception as e:
      self.logger.error('Could not find block definition for %s: %s' % (self.section, block, e)) 
      raise Exception('Could not find block definition.')
    # Check that we can find the block file.
    if blk_def == None:
      self.logger.error('Could not find block definition for %s.' % (self.section, block)) 
      raise Exception('Could not find block definition.')
    try:
      with open(blk_def) as f:
        f.close()
    except IOError:
      self.logger.error('Could not find block definition for \'%s\':\'%s\'. Suppose to be at \'%s\'')
      raise Exception('Could not read block definition.')
    except:
      raise
    print ("BLK DEF: %s" % blk_def)
    return blk_def
    
  def list_inputs(self, block):
    """Returns a list of host:file to check for before running the block.
       These are files which the block expects to be in place as input.
    """
    list = []
    if self.config[block]['input'].depth > 1:
      for input in self.config[block]['input'].keys():
        list.append(self.config[block][input])
    return list
  
  def list_outputs(self, block):
    """List the output files which should be created after running a block.
       If 'host' is not specified in host:file (ie. just 'file' is listed),
       the default of localhost should be assumed.
    """
    list = []
    if self.config[block]['output'].depth > 1:
      for output in self.config[block]['output'].keys():
        list.append(self.config[block][output])
    return list

  def load_block(self, block):
    """
    Loads and inspects the object for the block "section": "block" and then returns 
    the class instance for that block.
    """
    try: 
      blk_def = self.block_def_check(block)
    except Exception as e:
      self.logger.error('Failed to load block \'%s\':\'%s\': %s' % (self.section, block, e))
      raise Exception('Failed to load block: %s' % (e))
    
    class_inst = None
    
    # Is this a package we're pulling from? 
    # If we have a package (__init__.py exists), we want to add the path of the *package*
    # not the path of the file to our python import path.
    # Otherwise, we want to add the path of the file to our import path.
    path = os.path.dirname(blk_def)
    path_of_path = os.path.dirname(path)
    try:
      with open(path+'/__init__.py'):
        if path_of_path not in sys.path:
          sys.path.insert(0, path_of_path)
    except IOError:
      if path not in sys.path:
        sys.path.insert(0, path)
    
    mod_name, file_ext = os.path.splitext(os.path.split(blk_def)[-1])
    if file_ext.lower() == '.pyc':
      py_mod = imp.load_compiled(mod_name, blk_def)
    elif file_ext.lower() == '.py':
      py_mod = imp.load_source(mod_name, blk_def)
    else:
      self.logger.error('Definition for \"%s\" interface (%s) is not a py or pyc file.' % (block, blk_def))
      raise RuntimeError('Cannot load block.')
    
    # We are expecting the 'blockInterface' class, so check that we've found it.
    if not hasattr(py_mod, 'blockInterface'):
      self.logger.error('Definition for \'%s\' in \'%s\' lacks expected \'blockInterface\' class.' % (block, blk_def))
      raise RuntimeError('Connot load block.')
    try: 
      class_inst = py_mod.blockInterface()
    except Exception as e:
      self.logger.error('Could not load block \'%s\'. Failed to load \'blockInterface\' class: %s' % (block, e))
      raise e
    
    # Get the var/value pairs to set.
    var_value_pairs = self.get_var_value_pairs(block)
                                  
    # Get the required/expected vars that need to be set.
    # We look for the keyword 'required' and treat its value as a list of variable names
    # which need to be set.
    if hasattr(class_inst, 'REQUIRED'):
      for var in class_inst.REQUIRED:
        if var not in var_value_pairs:
          self.logger.error('Required value \'%s\' for block \'%s\':\'%s\' not found in configuration file.' % (var,self.section,block))
          raise('Cannot load block.')

    # Now set all the variables for our block's class instance.
    # NOTE: We'll reset values which contain strings that need interpolation (eg. %host%) later.
    for var, value in var_value_pairs.iteritems():
      if hasattr(class_inst, var):
        setattr(class_inst, var, value)
      else:
        self.logger.warn('Configuration file sets a value for \'%s\', but this argument is not used by the block \'%s\' as defined in \'%s\'.' % (var, block, blk_def))
    
    return class_inst
  
  def get_var_value_pairs(self, block):
    """
    Returns a dictionary of [var][value] pairs that act as args, in file names and out file names to a block.
    """
    var_value_pairs = {}
    for subsection in ['args', 'input', 'output']:
      if self.config[block][subsection] != None and self.config[block][subsection].depth > 1:
        for key in self.config[block][subsection].keys():
          var_value_pairs[key] = self.config[block][subsection][key]
    self.logger.info("Arguments for %s are: %s" % (block, var_value_pairs))
    return var_value_pairs
  
  def set_block_name_attr(self, block):
    """
    When we load a block, we'll add a NAME attribute,
    so hosts and block interfaces can refer to the block NAME
    when logging and reporting information.
    """
    if block not in self.blocks:
      self.logger.error('Asked to set NAME attribute on block \'%s\':\'%s\', but block interface is not loaded.' %(self.section, block))
      raise Exception('Could not set NAME.')
    try:
      setattr(self.blocks[block], 'NAME', block)
    except:
      raise
      
  def set_exp_dir_attr(self, block, exp_dir, exp_name):
    """
    When we load a block, we'll add a EXPDIR attribute,
    so hosts and block interfaces can refer to EXPDIR to determine
    where this experiment is located.
    """
    if block not in self.blocks:
      self.logger.error('Asked ot set EXPDIR attribute on block \'%s\':\'%s\', but block interface is not loaded.' % (self.section, block))
      raise Exception('Could not set EXPDIR and EXPNAME.')
    try:
      setattr(self.blocks[block], 'EXPDIR', exp_dir)
      setattr(self.blocks[block], 'EXPNAME', exp_name)
    except:
      raise
  
  def asks_for_block_target_attr(self, block):
    """
    Checks if a block requests the BLOCK_TARGET dictionary.
    This can be costly to set, so we only set this if
    the block requests it (by having the BLOCK_TARGET attribute).
    """
    # Do we have a block interface loaded for this block?
    if block not in self.blocks:
      self.logger.error('Asked to set BLOCK_TARGET attribute on block \'%s\':\'%s\', but block interface not loaded.' % (self.section, block))
      raise Exception('Could not set BLOCK_TARGET.')
    if hasattr(self.blocks[block], 'BLOCK_TARGET'):
      return True
    return False
  
  def set_block_target_attr(self, block, block_target_dict):
    """
    Just before running, we add a BLOCK_TARGET attribute,
    so blocks know who in the experiment is running what.
    This is important for blocks that do things like synchronization.
    """
    if self.asks_for_block_target_attr(block):
      try:
        setattr(self.blocks[block], 'BLOCK_TARGET', {})
        setattr(self.blocks[block], 'BLOCK_TARGET', block_target_dict)
      except:
        raise
  
  def set_group_attr(self, block, target_hosts):
    """
    Just before running, we add a GROUP attribute,
    so hosts and block interfaces can refer to GROUP to determine
    who else is currently running the same block software.
    """
    # Do we have a block interface loaded for this block?
    if block not in self.blocks:
      self.logger.error('Asked to set GROUP attribute on block \'%s\':\'%s\', but block interface not loaded.' % (self.section, block))
      raise Exception('Could not set GROUP.')
    try:
      setattr(self.blocks[block], 'GROUP', [])
      setattr(self.blocks[block], 'GROUP', target_hosts)
    except:
      raise
  
  def load_all_blocks(self,exp_dir='/tmp', exp_name='exp'):
    """
    Loads all blocks for a section ([extraction], [resource allocation] or [experiment])
    """
    # See if we have a call order. If so, we'll only pay attention to the blocks on the call order.
    try:
      order = self.block_call_order()
    except Exception as e:
      raise(e)
    
    self.blocks.clear()
    if order != None:
      for block in order:
        try: 
          self.blocks[block] = self.load_block(block)
          self.set_exp_dir_attr(block, exp_dir, exp_name)
          self.set_block_name_attr(block)
        except Exception as e:
          raise(e)
    else:
      for block in self.config.sections():
        try:
          self.blocks[block] = self.load_block(block)
          self.set_exp_dir_attr(block, exp_dir)
          self.set_block_name_attr(block)
        except Exception as e:
          raise(e)
    return 
    
  def get_depends_list(self, block, action):
    """ 
    Returns a list of blocks that a given block depends on (if any).
    """
    depends_list = []
    try:
      if self.config[block]['depends_on'] != None:
        for b in self.config[block]['depends_on']:
          if ':' in b:
            # Given action
            ba = b.split(':')[1]
            # Given block
            bb = b.split(':')[0]
            if ba == action:
              depends_list.append(bb)
          else:
            # Reguardless of the action, we depend on this block.
            depends_list.append(b)
              
        return depends_list
      return []
    except KeyError:
      return []
  
  def block_call_order(self):
    """
    Sanity checks and returns an ordered list of how blocks should be called.
    """
    call_list = []
    try:
      if self.config['order'] != None:
        for block in self.config['order']:
          try:
            test = self.config[block]
            call_list.append(block)
          except KeyError:
            self.logger.error('Section \'%s\' specifies a block called \'%s\' in \'order\', but this block is not defined in the section.' % (self.section,block))
            raise Exception('Unknown block.')
        return call_list
    except KeyError:
      self.logger.warn('Section \'%s\' does not specify block call order. Using order of defined blocks.' % (self.section))
      return self.config.sections
    return self.config.sections
  
  def blocks_to_force_kill(self, action):
    """
    Returns a list of blocks which are OK to force kill during the given action
    if we're waiting on these blocks.
    
    An example would be killing a monitoring tool when all other generation tools have quit
    """
    kill_list = []
    try:
      for block in self.block_call_order():
        if self.config[block]['force_quit_during'] != None:
          if action in self.config[block]['force_quit_during']:
            kill_list.append(block)
      print "Blocks to force kill this action:"
      print kill_list
      return kill_list      
    except KeyError:
      self.logger.warn('Section \'%s\' has no blocks which we can force kill (no set force_quit_during variable).' % (self.section))
      return []
      
  
  def action_order(self):
    """
    Reads from the config to get a list of actions and the order these actions 
    should be performed in. 
    Default is "install, setup, test, run, status".
    """
    action_list = []
    alist = []
    if self.config['actions'] != None:
      if not isinstance(self.config['actions'], list):
        alist = [self.config['actions']]
      else:
        alist.extend(self.config['actions'])
      for action in alist:
        action_list.append(action)
      return action_list
    else:
      # Return default.
      return ['install', 'setup', 'test', 'run', 'status']
  
  def interpolate_attr(self, block, interpolation={}):
    """
    Check for values which need to be interpolated, and replace
    the block's attributes which contain these interpolated strings.
    
    Eg. If an attribute called 'input_file' is set to 'input-%host%.txt', 
    we replace the %host% with whatever interpolation['%host%'] is.
    """
    if block not in self.blocks:
      self.logger.error('Asked to interpolate values on block \'%s\':\'%s\', but block interface not loaded.' % (self.section, block))
      raise RuntimeError('Could not interpolate.')
    
    var_value_pairs = self.get_var_value_pairs(block)
    for var, value in var_value_pairs.iteritems():
      # We check strings only
      if isinstance(value, basestring) and hasattr(self.blocks[block], var):
        for interpolated_str in interpolation:
          if value.contains(interpolated_str):
            newvalue = value.replace(interpolated_str, interpolation[interpolated_str])
            self.blocks[block].var = newvalue
          
  def action_function(self, block, action):
    """
    Runs the function for a specific section:block:action where section is 
    either [extraction], [resource allocation] or [experiment].
    """
    # Do we have a block interface loaded for this block?
    if block not in self.blocks:
      self.logger.error('Asked to call action \'%s\' on block \'%s\':\'%s\', but block interface not loaded.' % (action, self.section, block))
      raise Exception('Could not call action.')
    
    # Does the action exist in the block interface? If not, that's ok.
    if not hasattr(self.blocks[block], action):
      self.logger.warn('Asked to call action \'%s\' on block \'%s\':\'%s\', but this block interface does not define this action.' % (action, self.section, block))
      return     
    
    # Is the action callable (it better be).
    f = getattr(self.blocks[block], action)
    if not callable(f):
      self.logger.error('Asked to call action \'%s\' on block \'%s\':\'%s\', but action is not callable in block interface.' % (action, self.section, block))
      raise Exception('Could not call action.')
    print f  
    return f  

  def get_target(self, block):
    """
    For a given software block, we return the target (target hosts and groups), where
    the 'target' is the set of hosts the user wants to run the software block on.
    
    This function does not expand groups or differentiate between hosts and groups.
    That functionality is in the host handler and group handler.
    """
    targets = []
    if self.config[block]['target'] != None:
      for item in self.config[block]['target']:
        targets.append(item)
    return targets
      
      
      
      
      