from configobj import ConfigObj
from validate import Validator
import logging
import StringIO

import OrchestratorGlobals

class OrchestratorConfgHandler:
  """
  Reads and sanity checks the config file syntax.
  
  Breaks the config into parts for the Hosts, Groups and Block Handlers to use.
  """
  
  logger = None
  config = None
  
  configspec = StringIO.StringIO("""
  [logging]
    log_file = string(max=80, default='./orchestrator.log')
    log_level = integer(0,2, default=2)
    exp_dir = string(max=80, default='/tmp/')
    exp_name = string(max=80, default='exp')
  
  [hosts]
    user = string(max=255, default=None)
    gateway = string(max=255, default=None)
    key_filename = force_list(default=None)
    reject_unknown_hosts = boolean(default=False)
    shell = string(max=255, default=None)
    use_shell = boolean(default=False)
    ssh_config_path = string(max=255, default=None)
    timeout = integer(default=10)
    warn_only = boolean(default=True)
    nodes = force_list(default=None)
    sudo = boolean(default=True)
    tmux = string(max=255, default='/usr/bin/tmux')
    
    [[__many__]]
      user = string(max=255, default=None)
      gateway = string(max=255, default=None)
      key_filename = force_list()
      reject_unknown_hosts = boolean(default=False)
      shell = string(max=255, default=None)
      use_shell = boolean(default=False)
      ssh_config_path = string(max=255, default=None)
      timeout = integer(default=10)
      warn_only = boolean(default=True)
      nodes = force_list()
      sudo = boolean(default=True)
      tmux = string(max=255, default='/usr/bin/tmux')
    
  [groups]
  
  [extraction]
    order = force_list(default=None)
    install_order = force_list(default=None)
    setup_order = force_list(default=None)
    run_order = force_list(default=None)
    cleanup_order = force_list(default=None)
    
    [[__many__]]
    def = string(max=80)
    target = force_list(default=None)
    force_quit_during = force_list(default=None)
    depends_on = force_list(default=None)
    [[[input]]]
    [[[output]]]
    [[[args]]]

  [resource allocation]
    order = force_list(default=None)
    install_order = force_list(default=None)
    setup_order = force_list(default=None)
    run_order = force_list(default=None)
    cleanup_order = force_list(default=None)
    
    [[__many__]]
    def = string(max=80)
    target = force_list(default=None)
    mapping = force_list(default=None)
    force_quit_during = force_list(default=None)
    depends_on = force_list(default=None)
    [[[input]]]
    [[[output]]]
    [[[args]]]

  [experiment]
    order = force_list(default=None)
    install_order = force_list(default=None)
    setup_order = force_list(default=None)
    run_order = force_list(default=None)
    cleanup_order = force_list(default=None)
    
    [[__many__]]
    def = string(max=80)
    target = force_list(default=None)
    force_quit_during = force_list(default=None)
    depends_on = force_list(default=None)
    [[[input]]]
    [[[output]]]
    [[[args]]]
  
  """)
  
  def __init__(self, config_file, level=None):
    """
    Reads the config file and checks syntax.
    
    Throws IOError and ConfigObj exceptions.
    """
    # Check if we can open the config
    try:
      with open(config_file) as f: pass
    except IOError as e:
      raise e
    except Exception as e:
      raise e
    
    # Check if the config file has correct syntax.
    try: 
      self.config = ConfigObj(config_file, configspec=self.configspec, file_error=False, interpolation='Template')
    except:
      raise
    val = Validator()
    test = self.config.validate(val)
    if test == False:
      raise Exception('Config file contains errors.')
    else:
      print 'No errors found in configuration file: %s' % config_file

    # Check out the logging config, and set up the logger.
    try:
      self.set_logger(level=level)
    except:
      raise

    
  def set_logger(self, level=None):
    """
    Sets up and configures logging.
    """
    file = self.config['logging']['log_file']
    if file != None:
      try:
        if level == None:
          lvl = int(self.config['logging']['log_level'])
        else:
          try:
            lvl = int(level)
          except Exception as e:
            raise Exception('Given invalid value for log level: %s' % (e))
        
        self.logger = logging.getLogger('Orchestrator')
        self.logger.setLevel(logging.DEBUG)
        self.propagate = True
        
        #fh = logging.FileHandler(file)
        #fh.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)
        
        #formatter = logging.Formatter('%(levelname)s %(asctime)s: %(message)s')
        #fh.setFormatter(formatter)
        
        #self.logger.addHandler(fh)
        self.logger.addHandler(ch)
        
        # We're using a catch-all logger so we get fabric logs.
        logging.basicConfig(format='%(levelname)s %(asctime)s: %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p', filename=self.config['logging']['log_file'], level=lvl, filemode='a+') 
      except Exception as e:
        raise e
    else:
      raise Exception('Not able to log.')
  
  def return_exp_dir(self):
    return self.config['logging']['exp_dir']
  
  def return_exp_name(self):
    return self.config['logging']['exp_name']
  
  def return_logger(self):
    return self.logger 
  
  def return_hosts_config(self):
    return self.config['hosts']
  
  def return_group_config(self):
    return self.config['groups']
  
  def return_block_config(self, section):
    if section in OrchestratorGlobals.ALLOWED_SECTIONS:
      return self.config[section]
    else:
      return None
  