from fabric.api import hosts, run, task, env, settings, hide, execute
import logging
import OrchestratorGlobals

class OrchestratorDetachHandler:
  session_name = ""

  def __init__(self, session_name):
    """
    Creates a DetachHandler for an experiment with name 'session_name'.
    """
    self.session_name = session_name
 
  def start_session(self):
    tmux_string = '%s new-session -d -s %s' % (env.tmux, self.session_name)
    try:
      return run(tmux_string)
    except:
      raise
    
  def clean_up(self):
    """
    Used on ctrl-C, or another unexpected shutdown event.
    Kills current session, all windows attached to it and any clients connected.
    """
    tmux_string = '%s kill-session -t %s' % (env.tmux, self.session_name)
    try:
      return run(tmux_string)
    except:
      raise
  
  def window_exists(self, window):
    window_name = window.split(':')[1]
    tmux_string = '%s list-windows -t %s' % (env.tmux, self.session_name)
    try:
      list = run(tmux_string).split()
      print list
      if window_name in list:
        return True
      else:
        return False
    except:
      raise

  def new_window(self):
    window_name = env.tmux_window.split(':')[1]
    tmux_string = '%s new-window -n %s -t %s' % (env.tmux, window_name, self.session_name)
    print tmux_string
    try:
      return run(tmux_string)
    except:
      raise
  
  def queue_window_for_close(self, target=''):
    tmux_string = "%s send-keys -t %s \"tmux kill-window -t %s\" C-m" % (env.tmux, target, target)
    try:
      return run(tmux_string)
    except:
      raise
  
  def list_windows(self):
    tmux_string = "%s list-windows -t %s -F ':#{window_name}:'" % (env.tmux, self.session_name)
    try:
      return run(tmux_string)
    except:
      raise
      
  def check_for_session(self):
    """
    Returns True if session_name exists as a tmux session.
    """
    try:
      output = execute(self.tmux_list_sessions)
    except:
      raise
    # If we're running fabric in task mode, execute will return
    # a dictionary of {host:response} pairs. If this is called on one host,
    # we'll just get the output.
    name = self.session_name
    if isinstance(output, dict):
      for o in output:
        if name not in output[o]:
          return False
      return True
    else:
      if name not in output:	
        return False
      else:
        return True
   
  def tmux_start_new_session(self):
    tmux_string = '%s new-session -d -s %s' % (env.tmux, self.session_name)
    try:
      return run(tmux_string)
    except:
      raise
     
  def tmux_list_sessions(self):
    tmux_string = "%s list-sessions" % (env.tmux)
    try:
      return run(tmux_string)
    except:
      raise
  
  def tmux_kill_session(self):
    tmux_string = '%s kill-session -t %s' % (env.tmux, self.session_name)
    return run(tmux_string)
  
  def tmux_new_window(self):
    tmux_string = '%s new-window -n %s -t %s' % (env.tmux, env.block_name, self.session_name)
    return run(tmux_string) 
    
  def attach(self):
    """
    Attach to an existing session. Error out if we can't find it. 
    """
  
  def detach(self):
    """
    Detach from an existing session. Error if we can't find it.
    """
    
  def is_running(self, process_name):
    """
    Checks the status of a process in a session.
    We check by looking for a window name in the session, 
     and then check what's going on in that window.
    """
    
  def peek(process_name):
    """
    Returns the last few lines of output from a process in a session.
    """
  
    
  def transfer(self, filename, from_name, to_name, from_host = None, to_host = None, sudo=True):
    """
    Transfer a file from a location to a location. The location may have a destination host and source host.
    If none is provided, the from or to_host is the local host.
    Returns True if no errors were reported.
    Returns False if an error occurs.
    """
    return True
  
        
class OrchestratorFunctions:
  def run(self, command, sudo=True, second_attempt=False):
    """
    Run command on target machine. Is run as sudo by default.
    Checks if tmux session exists (starts one if not),
    checks if window exists (starts one if not),
    checks if any process is running in window already
    runs process in window and detaches.
    
    Returns output from running tmux command.
    
    Gets 'exp_name' from the env, and uses this name to find the tmux session.
    Uses the 'block_name' (given by the software block at run time, 
    usually from the self.NAME variable dynamically added to the software block
    at run time) to choose/create a window in the tmux session.
    
    Checks that the tmux session exists and window exists (if they don't, we raise an error).
    Checks if the window is currently busy, 
    (if so, we wait for up to env.max_wait seconds, trying to start our 'command')
    
    If we wait env.max_wait seconds and still the window is busy, we give up 
    and raise an error. XXX At some point this should be an Orchestrator Error Class error.
    """
    try:
      logger = logging.getLogger('Orchestrator')
    except Exception as e:
      raise(e)
    try:
      if env.host == None:
        h = env.host_string
      else:
        h = env.host
      print("Command is %s on %s"%(command, h))
    except:
      raise Exception("Called without access to host in global env")
      
    # If we're not using tmux, we can just run command normally.
    if env.tmux == 'None':
      print "Running command without tmux"
      logger.info("Running %s on %s without detach session (tmux)." % (command, h))
      try:
        return run(command)
      except:
        raise
    
    # Else, wrap command in tmux. When called from Orchestra,
    # the tmux active window is a new window created for this block:function.
    else:
      print "Running with detach in tmux" 

      # If command has quotes (") then we should just run it normally.
      if '"' in command:
        logger.warning("Command %s contains '\"'. Currently do not support running in detach handler (tmux)" % (command))
        try:
          return run(command)
        except:
          raise

      # Add command to list for active window
      try:
        tmux_window = env.tmux_window
        exp_name = tmux_window.split(':')[0]
      except AttributeError:
        print "Cannot determine detach handler session:window to add command. Running without detach."
        logger.warning("Could not determine detach handler session:window. Running command without detach on %s." %(h))
        try:
          return run(command)
        except:
          raise
      detachHandler = OrchestratorDetachHandler(exp_name)
      
      # Check for window, start it if it doesn't exist.
      if not detachHandler.window_exists(env.tmux_window):
        detachHandler.new_window()
      
      wrapped_command = "%s send-keys -t %s \"%s\" C-m" % (env.tmux, env.tmux_window, command)
      print("Running %s on %s" % (wrapped_command, h))
      logger.info("Running %s on %s" % (wrapped_command, h))    
      try:
        return run(wrapped_command)      
      except NetworkError, err:
        if not second_attempt:
          self.run(command, sudo, second_attempt=True)
        else:
          raise
      except ssh.SSHException, err:
        if not second_attempt:
          self.run(command, sudo, second_attempt=True)
        else:
          raise
      except:
        raise
    
  def my_part(self, group_list):
    """
    Takes a group list and using the env.host or env.host_string variable (from global env)
    determines which part number this host is. 
    
    The GROUP list is always ordered alphabetically, so throughout the experiment,
    a host will always have the same part number within a given group.
    
    Returns a string (str(int))
    """
    if env.host == None:
      host = env.host_string
    else:
      host = env.host
    i = 0
    for h in group_list:
      if host == h:
        break
      i = i+1
    return str(i)
      