import logging
from configobj import ConfigObj

class OrchestratorGroupHandler:
  """
    Handles the parsing, testing and expansion of groups.
    
    Parses the "[groups]" section of the config file.
  """
  group_config = None
  groups = []
  group_dict = {}
  logger = None
  max_nesting_depth = 4
  
  def __init__(self, group_config, max_nesting_depth = 4):
    self.group_config = group_config
    self.logger = logging.getLogger('Orchestrator')
    self.max_nesting_depth = max_nesting_depth
    
    # Fist, get a fixed list of group names.
    for group in self.group_config.scalars:
      if group in self.groups:
        raise Exception('Multiple definitions for group %s.' % (group))
      self.groups.append(group)
          
  def is_group(self, entity):
    if entity in self.groups:
      return True
    return False
  
  def entity_members(self, group):
    """
    Returns a list of entities (groups and hosts) defined as the given group in the config or
    null if the name given is not a group defined in the config.
    """
    if not self.is_group(group):
      raise RuntimeError('Asked about group %s. Group not defined in config.'  % (group))
      return None
      
    # configObj won't return a standard python list necessarily, so we make sure
    # that we return a list here, not just a configObj.
    members = []
    if isinstance(self.group_config[group], basestring):
      # We may get a singlet returned (ie. not a list, just a single entity), but all
      # functions expect a group definition to be a list of items, so we need to force
      # a list.
      members.append(self.group_config[group])
    else:
      # We have a list returned from the config file.
      for m in self.group_config[group]:
        members.append(m)
    # Remove any duplicates from the list members.
    return list(set(members))
    
  def expand_group(self, group_name, recursion_depth = 0):
    """
    Given a list of entities, expands all groups and returns a *set* of hosts. 
    
    Hostnames are what's listed in the config file, not the full and checked hostname
    given by the OrchestratorHostHandler class, so hostnames returned may not be real hostnames.
    All names should be checked with the OrchestratorHostHandler class before use.
    
    Groups can be nested up to a depth of max_nesting_depth (default 4 levels). 
    """
    
    expanded_group_list = []
    group_list = self.entity_members(group_name)
    
    # We recurrsively expand the groups in the list, up until we have no more groups
    # in the list, or we've hit our nested group limit.
    for member in group_list:
      # Group: if an entity is a group, we recursively expand it..assuming we haven't hit our
      # max nesting depth.
      if self.is_group(member):
        if recursion_depth < self.max_nesting_depth:
          for submember in self.expand_group(member, recursion_depth + 1):
            expanded_group_list.append(submember)
        else:
          self.logger.warn('Group %s not further expanded. Nesting depth exceeded.' % (member))
      # Host: if an entity is not a group, we assume it's a host and does not need to be expanded.
      else:
        # This member isn't a group, so we assume it's a host and add it.
        expanded_group_list.append(member)
    
    # Remove duplicates.
    expanded_group_list = list(set(expanded_group_list))
    return list(set(expanded_group_list))
  
  def host_members(self, group):
    """
    Given a group name, returns all the hosts which belong to the group.
    (Does a list_group and expand_group)
    """
    return self.expand_group(group)
    
  def member_number(self, group, entity):
    """ 
    Given an entity, returns the member number of that entity.
    
    Useful for auto-assigning files to hosts based off numbers.
    Eg. a trace split into 10 parts. Each part trace file can be numbered,
    and will automatically get assigned to a host.
    """
    i = 0
    for member in self.host_members(group):
      if member == entity:
        break
      i = i+1
    return i
    
    