import getpass
import signal
import inspect
import copy
import subprocess
import os
import re
import logging
from configobj import ConfigObj
from fabric.api import hosts, run, sudo, task, env, settings, hide, execute
from fabric.operations import get, put
from fabric.state import commands, connections
from fabric.contrib.files import exists
from fabric.network import ssh
from fabric.exceptions import NetworkError

import OrchestratorGlobals
import OrchestratorFunctions

class HostInfo:
  keepalive = 10
  user = None
  gateway = None
  key_filename = None
  reject_unknown_hosts = False
  use_shell = False
  shell = None
  ssh_config_path = None
  warn_only = True
  timeout = 30
  sudo = True
  tmux = '/usr/bin/tmux'
  perl = '/usr/bin/perl'
  python32 = '/usr/bin/python3.2'
  python27 = '/usr/bin/python'

def run_with_retry(host_list):
  def retry_fabric_func(func):
    def call_and_retry(*args, **kwargs):
      logger = logging.getLogger('Orchestrator')
      if not isinstance(host_list, list):
        first = host_list
      else:
        if len(host_list) < 1:
          print("ERROR: Not given target list of hosts to run on?")
          return
        first = host_list[0]
      try:
        print("Trying to run function with wrapper")
        ret = func(*args, **kwargs)
        if isinstance(ret, dict):
          for host in ret:
            if isinstance(ret[host], NetworkError):
              logger.warning("Ran into trouble with connection (%s). Restarting conn." %(host))
              print("WARNING: Ran into trouble with connection (%s). Restarting connection." %(host))
              try:
                connections[host].close()
                del connections[host]
              except Exception as e:
                print("Failed to close connection. (%s)" % e)
                pass
              try:
                with settings(hosts=host):
                  f = run_with_retry(host)(func)
                  ret = f(*args, **kwargs)
                  return ret
              except Exception as e:
                print("Second (2nd) try failed")
                raise(e)
        return ret
      except NetworkError, err:
        pass
      except ssh.SSHException, err:
        pass
      except Exception as e:
        print("ERRROR: (in wrapper) %s" %e)
        raise(e)
      print("WARNING: Ran into trouble with connection (%s). Restarting connection." %(first))
      logger.warning("Ran into trouble with connection (%s). Restarting conn." %(first))
      try:
        connections[first].close()
        del connections[first]
      except Exception as e:
       print("WARNING: Problem closing bad connection to %s: %s" % (first, e))
       pass
      print("WARNING: Problem with fabric connection. Trying again.")
      try:
        ret = func(*args, **kwargs)
        return ret
      except Exception as e:
        print("Second try failed. %s" % (e))
        raise(e)
    return call_and_retry
  return retry_fabric_func

class OrchestratorHostsHandler:
  """
  Handles interpreting the config file's "[node]" section which specifies the target hosts 
  which will run the experiment.
  
  The "[node]" section has the following possible global variables which will be used
  to connect to all hosts unless otherwise specified in a subsection dedicated to a domain/hostname:
    user (ie. an ssh login name)
    gateway  (optional hostname of a gateway machine. For example, users.isi.deterlab.net is a gateway host for all of deter)
    key_filename (SSH key file to use when connecting)
    reject_unknown_hosts (Refuse to connect to hosts which are not in the known hosts file. True or False)
    shell (defaults to /bin/bash)
    ssh_config_path (Use some other config other than ~/.ssh/config)
    timeout (Network connection timeout, in seconds. Default is 10s)
    hosts (List of short or long hostnames to be used in the experiment.)
    
  All subsections in the "[node]" section should be domains for which the above variables
  need to be different. For example, to use hosts in the .mydomain.net domain 
  which has a different username login than my other hosts, define the following section:
  
    [mydomain.net]
      user = mylogin
      key_filename = ~/mykeyfile
      hosts = host1, host2
  
  Given such a section, Orchestrator will log into "host1.mydomain.net" as "mylogin"
  and use the ssh key found in ~/mykeyfile.
  
  """

  localhost_strings = ["localhost", "127.0.0.1", "local"]
  host_config = None
  
  def __init__(self, host_config):
    self.logger = logging.getLogger('Orchestrator')
    self.host_config = host_config
    self.host_dict = {}
    try:
      self.expand_hosts()
    except Exception as e:
      raise(e)
  
  def expand_hosts(self):
    """
    Expands the information in the "[hosts]" config section into a dictionary
    of hostsnames and environments.
    """
    # Get our global setting for hosts.
    global_settings = HostInfo()
    
    # Go through the settings in the global section.
    for setting in self.host_config.scalars:
      if hasattr(HostInfo, setting):
        setattr(global_settings, setting, self.host_config[setting])
    
    # We'll always set a user name, so if the global settings don't sepcify a user name
    # we'll use getpass to get one.
    if global_settings.user == None:
      global_settings.user = getpass.getuser()
    
    # First look at the hosts in the global section.
    if self.host_config['nodes'] != None:
      for host in self.host_config['nodes']:
        if host in self.host_dict:
          raise RuntimeError('Multiple definitions for some host.')
        self.host_dict[host] = copy.copy(global_settings)

    # Go through subsections. Subsection = domain
    for domain in self.host_config.sections:
      if self.host_config[domain]['nodes'] != None:
        for host in self.host_config[domain]['nodes']:
          if host in self.host_dict:
            raise RuntimeError('Multiple definitions for some host.')
          # XXX Probably going to be problematic... we need a string, we have a section.
          # First put all the global setting in for the host as a default.
          self.host_dict[self.add_name(host,domain)] = copy.copy(global_settings)
          # Now adjust the setting for this host to the settings defined in the subsections.
          setattr(self.host_dict[self.add_name(host,domain)], 'host_string', self.add_name(host, domain))
          for setting in self.host_config[domain].scalars:
            if hasattr(HostInfo, setting) and self.host_config[domain][setting] != None:
              setattr(self.host_dict[self.add_name(host,domain)], setting, self.host_config[domain][setting])
      
    self.logger.info("Experiment is set to use these hosts: %s" % (self.list_hosts()))
  
  def add_name(self,host,domain):
    """
    Given a host name (such as myhost or myhost.mysubnet) and a domain (.net, .isi.edu etc.)
    combines the two to make a full hostname.
    """
    return host.strip('.') + '.' + domain.strip('.')
    
  def full_name(self, nickname):
    """
    If we can, return a full host name for the nick name.
    
    If nickname may be two different hosts, then we return None.
    
    If a nickname does not seem to be a host we know about, we return None.
    """
    matches = []
    split_short = nickname.split('.')
    for full_hostname in self.host_dict:
      split_full = full_hostname.split('.')
      i = 0
      a_match = True
      if len(split_short) <= len(split_full):
        for part in split_short:
          if part != split_full[i]:
            a_match = False
            break
          i = i+1
        if a_match == True:
          matches.append(full_hostname)
    if len(matches) == 0:
      return None    
    if len(matches) > 1:
      self.logger.warn('Short hostname \'%s\' could be one of multiple hosts: %s' % (nickname, matches))
      return None
    else:
      return matches[0]
            
  def list_hosts(self):
    """
    Parses the config file's "[node]" section and returns a list 
    of the fullnames of all hosts (if possible).
    """
    return list(self.host_dict.keys())
  
  def is_host(self, target):
    """
    Throughout the experiment, we need to check that the target of a task is actually a host
    we know how to get to.
    
    Returns True or False
    """
    if target in self.host_dict:
      return True
    return False

  def clean_up(self, exp_name):
    """
    Cleans up any state on hosts in the event of an exit.
    """
    dH = OrchestratorFunctions.OrchestratorDetachHandler(exp_name)
    
    for host in self.list_hosts():
      print("Cleaning up state on host %s." % host)
      try:
        env_kwargs = self.host_env(host)
        if env_kwargs['tmux'] != 'None':    
          env_kwargs['parallel'] = False
          env_kwargs['warn_only'] = True
          self.logger.info("Cleaning up on host %s" % (host))
          with settings(**env_kwargs), hide('output', 'running', 'warnings'):
            execute(dH.clean_up)
      except (KeyboardInterrupt, SystemExit):
        pass  
      except Exception as e:
        raise(e)
  
  def install_tmux(self):
    try:
      for host in self.list_hosts():
        env_kwargs = self.host_env(host)
        if env_kwargs['tmux'] != None:
          print("Installing tmux on %s" % host)
          with settings(**env_kwargs), hide('output', 'running', 'warnings'):
            f = run_with_retry(host)(sudo)
            f('apt-get -y install tmux')
    except Exception as e:
      print("Problem installing tmux: %s" % e)
      raise
    return True

  def check_hosts(self):
    """
    Goes through the list of hosts in the "[node]" section and attempts login.
    """
    ok_flag = True
    for host in self.list_hosts():
      env_kwargs = self.host_env(host)        
      env_kwargs['parallel'] = False
      env_kwargs['warn_only'] = True
      env_kwargs['sudo_user'] = 'root'
      env_kwargs['allow_agent'] = False

      self.logger.info("Collecting passwords and checking connectivity to %s" % (host))
      print("\033[1m%s\033[0m:" %(host))
      
      # If we have a gateway to get through, check
      # that this host really does exist on the other side.
      # Otherwise, fabric will just keep prompting the user for
      # a password with no other feedback.
      if 'gateway' in env_kwargs:
        print("Testing credentials to gateway for %s (%s)" %(host, env_kwargs['gateway']))
        self.logger.info("Testing gateway (%s) used for %s" % (env_kwargs['gateway'], host))
        gateway_env = {}
        for setting in env_kwargs:
          gateway_env[setting] = env_kwargs[setting]
        # Remove the gateway.
        gateway_env['gateway'] = None
        gateway_env['host_string'] = env_kwargs['gateway']
        # Attempt to see if the remote host exists.
        with settings(**gateway_env), hide('output', 'running', 'warnings'):
          result = run('/sbin/ping -c 2 %s' % host)
          if result.return_code != 0:
            self.logger.warn('Was not able to reach host %s from gateway %s' % (host, gateway_env['host_string']))
            raise Exception('Was not able to reach host %s from gateway %s' % (host, gateway_env['host_string']))
        print("Connectivity to %s via gateway looks ok." % (host))
        self.logger.info("Connectivity to %s via gateway looks ok." % (host))
      
      with settings(**env_kwargs), hide('output', 'running', 'warnings'):
        try:
          if env_kwargs['sudo'] == True:
            f = run_with_retry(host)(sudo)
            f('hostname')
            #XXX result = sudo('hostname')
            if result.failed:
              self.logger.warn("Problem running basic sudo command on %s... might be ok?" %(host))
              f = run_with_retry(host)(run)
              result = f('hostname')
              #result = run('hostname')
          else:
            f = run_with_retry(host)(run)
            result = f('hostname')
          if result.failed:
            self.logger.warn("Problem running basic command (\'hostname\') on %s" %(host))
            ok_flag = False
        except Exception as e:
          # Uh-oh, can't reach the machine.
          self.logger.warn("Problem logging into %s: %s" % (host, e))
          print("Problem logging into %s:%s" % (host, e))
          ok_flag = False
      self.logger.info("Done checking connectivity for %s" % (host))
      print("Done checking connectivity for %s\n" %(host))
   
    return ok_flag
  
  def setup_detach_handler(self, name):
    """
    Sets up a detach handler on the target machine (ie. starts a tmux session).
    Returns True if all hosts were set up as expected.
    Returns False if a host is marked to run a detach handler, but can't.
    """
    self.logger.info("Setting up detach (tmux) sessions on hosts.")
    detachHandler = OrchestratorFunctions.OrchestratorDetachHandler(name)
    for host in self.list_hosts():
      env_kwargs = self.host_env(host)
      env_kwargs['parallel'] = False
      env_kwargs['warn_only'] = True
      if env_kwargs['tmux'] != 'None':
        with settings(**env_kwargs), hide('output', 'running', 'warnings'):
          if not detachHandler.check_for_session():
            self.logger.info("Starting detach (tmux) session on %s" % (host))
            print("Starting detach (tmux) session for %s" % (host))
            result = detachHandler.start_session() 
            if result.failed:
              self.logger.error("Could not set session for %s using %s: %s" % (host,env_kwargs['tmux'],result))
              return False
          else:
            self.logger.info("Detach handler already running on %s" % (host))
            print("%s already running detach handler" % (host))
      else:
        self.logger.info("Host %s will not have a detach handler set up." % host)
    return True 
  
  def host_env(self, target):
    """
    Returns a string of the login environment 
    (eg. full host name, gateway, login etc.) for a particular host target.
    """
    if not self.is_host(target):
      raise RuntimeError('Asked for information on unknown target.')
      
    env_kwargs = {}
    
    for attr, value in inspect.getmembers(self.host_dict[target]):
      # A bit hacky, but we're skipping the default functions
      # and only getting the attributes of our HostInfo class.
      if not attr.startswith('__'):    
        if value != None:
          env_kwargs[attr] = value
      
      # For some reason we regularly lose the ssh-agent thread
      # in paramiko... causing everything to fail. So don't use the agent!
      env_kwargs['no_agent'] = True  
    return env_kwargs

  def run_on_host(self, host_list, function, exp_name, block_name, action, second_try = False):
    # First, we have to sort the host_list into groups we 
    # can call in parallel - ie. groups which have the same
    # environment.
    dH = OrchestratorFunctions.OrchestratorDetachHandler(exp_name)
    try:
      groups = self.group_hosts_by_env(host_list)
      for group in groups:
        env_kwargs = self.host_env(group[0])
        if len(group) > 1:
          env_kwargs['parallel'] = True
          env_kwargs['pool_size'] = 10
          env_kwargs['host_string'] = ''
          env_kwargs['tmux_window'] = exp_name + ':' + block_name + '-' + action
          self.logger.info("Running task on multiple machines in parallel. Running on: %s" %(group))
          with settings(**env_kwargs), hide('output', 'running', 'warnings'):	
            # If we're using detach via tmux, open window for this function and make window active.
            #if env_kwargs['tmux'] != 'None':
            #  print("Starting new window")
            #  self.logger.info("Starting new window %s on %s" % (env_kwargs['tmux_window'].split(':')[1], group))
            #  try:
            #    f = run_with_retry(group)(execute)
            #    result = f(dH.new_window, hosts=group)
                #result = execute(dH.new_window, hosts=group)
            #  except Exception as err:
            #    self.logger.error("Problem starting tmux window: %s" % err)
            #    raise(err)
          
            # Run the function. Any calls to OrchestratorFunctions.run() will use tmux
            # to run detached if possible.
            self.logger.info("Running %s function on group of hosts." % block_name)
            try:
              f = run_with_retry(group)(execute)
              output = f(function, hosts=group)
              #output = execute(function, hosts=group)
            except Exception as err:
              self.logger.error("Problem running function %s: %s" % (block_name, err))
              raise(err)
          
            # If we're using detach via tmux, send signal to close window when done.
            if env_kwargs['tmux'] != 'None':
              self.logger.info("Queuing window for close %s" % env_kwargs['tmux_window'].split(':')[1])
              try:
                f = run_with_retry(group)(execute)
                f(dH.queue_window_for_close, target=env_kwargs['tmux_window'], hosts=group)
                #execute(dH.queue_window_for_close, target=env_kwargs['tmux_window'], hosts=group)
              except Exception as err:
                self.logger.error("Problem sending close command for window. Block %s. %s" % (block_name, err))
                raise(err)
        else:
          env_kwargs['tmux_window'] = exp_name + ':' + block_name + '-' + action
          self.logger.info("Was not able to group hosts to run task in parallel. Running on single host %s" %(group))
          with settings(**env_kwargs), hide('output', 'running', 'warnings'):
            f = run_with_retry(group)(execute)
            # If we're using detach via tmux, open window for this function and make window active.
            if env_kwargs['tmux'] != 'None':
              self.logger.info("Starting new window %s" % env_kwargs['tmux_window'].split(':')[1])
              try:
                result = f(dH.new_window)
              except Exception as err:
                self.logger.error("Problem opening new tmux window on %s" % group)
                raise(err)
            try:
              f(function)
            except Exception as err:
              self.logger.error("Problem running: %s" % err)
              raise(err)

            # If we're using detach via tmux, send signal to close window when done.
            if env_kwargs['tmux'] != 'None':
              self.logger.info("Queuing window for close %s" % env_kwargs['tmux_window'].split(':')[1])
              try:
                f(dH.queue_window_for_close, target=env_kwargs['tmux_window'])
              except Exception as err:
                self.logger.error("Problem closing window on %s" % group)
                raise(err)

    except ssh.SSHException, err:
      if str(err) == 'SSH session not active' and not second_try:
        print("WARNING: Problem with ssh connection. Trying again.")
        connections[host_list[0]].close()
        del connections[host_list[0]]
        self.logger.error("Problem running function. Connection close to host %s. Trying again." % (host_list))
        self.run_on_host(host_list, function, exp_name, block_name, stage_name, second_try = True)
      else:
        self.logger.error("Problem running function on %s: %s" %(host_list, e))
        raise(e)
    except NetworkError, err: 
      if str(err) == 'Error reading SSH protocol banner' and not second_try:
        connections[host_list[0]].close()
        del connections[host_list[0]]
        print("WARNING: Problem with ssh banner. Trying again.")
        self.logger.error("Problem running function. Problem with ssh banner on %s. Trying again." % (host_list))
        self.run_on_host(host_list, function, exp_name, block_name, stage_name, second_try = True)
    except Exception as e:
      self.logger.error("Problem running function on %s: %s" % (host_list, e))
      raise(e)
  
  def action_not_done(self, action, exclusion_list, exp_name, block='', second_try=False):
    """
    Returns True if actions are not done. Excludes blocks listed in the exclusion_list from
    concideration.
    """
    exclude_window_names = []
    for e in exclusion_list:
      exclude_window_names.append(e + '-' + action)
      
    # Get a list of windows in the existing tmux session on a host.
    existing_windows = {}
    host_list = self.list_hosts()
    dH = OrchestratorFunctions.OrchestratorDetachHandler(exp_name)
    try:
      for host in host_list:
        env_kwargs = self.host_env(host)
        if env_kwargs['tmux'] != 'None':
          with settings(**env_kwargs), hide('output', 'running', 'warnings'):
            output = execute(dH.list_windows)
          windows = ''
          if isinstance(output, dict):
            for o in output:
              windows = windows + ':' + output[o]
          else:
            windows = output
          window_list = re.split('\n|\:', windows)
          task_to_look_for = block + '-' + action
          for w in window_list:
            if w not in exclude_window_names and task_to_look_for in w:
              self.logger.info("Waiting on %s on %s (and possibly others)." % (w, host))
              print("Waiting on %s on %s (and possibly others)." % (w, host))
              return True
    except ssh.SSHException, err:
      self.logger.error("Problem checking state of host %s. Connection dead." %(host_list))
      print("ERROR: Problem checking running state of host. Connection dead. Re-try not implemented.")
      return True
    except Exception as e:
      self.logger.error("Problem checking state of host %s." % (e))
      print("ERROR: Problem checking state of host: %s" % (e))
      return False
    return False

  def copy_files(self, host, var_value_pairs, interpolation_dict, type='output'):
    """
    Take a list of inputs or outputs (var_value_pairs) for a given host 
      (produced or consumed by a particular block) 
      
    Each item in the list should have the format:
    /complete/path/to/file, host:/full/path [...] host:/full/path   
    
    In inputs, the /complete/path/to/file is the location on the host
    a file should be copied and the following host:path pairs are locations to get the file.
    
    In outputs, the /complete/path/to/file is the location the file on the host, 
    and it should be copied to the folling host:path locations (if given).
    
    The interpolation dictionary gives %variable% value pairs to replace
    in the files. Eg. replace '%host%' with the host name.
    """
    
    if not self.is_host(self.full_name(host)):
      self.logger.error("Given host we don't know about. (%s)" %(host))
      return
      
    host_env = self.host_env(host)  
    for var, val in var_value_pairs.iteritems():
      # Get the source/dest
      if isinstance(val, list):
        file = str(val[0])
      else:
        if ',' in val:
          file = val.split(',')[0].strip()
        else:
          file = val.strip()
          self.logger.debug('Given %s file (%s), but no location to source/dest this resource.' % (type, val))
          
      if not '/' in file:
        self.logger.warn('Unclear if given a true path (%s) for %s.' %(file, type))
      
      # If this is output, check that this file was produced.
      if type == 'output':        
        with settings(**host_env), hide('output', 'running', 'warnings'):
          if not exists(file):
            self.logger.error('Output file %s was not created on host %s.' %(file, host))
            continue

      if isinstance(val, list): 	
        if len(val) > 1:
          resources = val[1:]
      else:
        tmp = val.split(',')
        if len(tmp) > 1:
          resources = tmp[1:]
        else:
          resources = []

      # For now, we check the first resource. 
      for resource in resources[0:1]:
        h = ''
        if ':' in resource:
          h = resource.split(':')[0]
          if h == '':
            h = 'localhost'
          else:
            if not self.is_host(self.full_name(h)):
              self.logger.error('Given hostname %s for %s, but is not a known hosts.' % (h, type)) 
              raise Exception('Could not parse %s variable. Unknown host.' % (type))
          try:
            p = resource.split(':')[1]
          except:
            self.logger.error('Given hostname %s for %s, but no path name.' %(h, type))
            raise Exception('Could not parse %s variable. No path given.' %(type))
        else:
          p = resource.strip()
        if not '/' in p:
          self.logger.warn('Unclear if given path (%s) for %s is valid.' %(p, type))
        # Interpolate on resource file/path:
        if '%' in p:
          for key in interpolation_dict:
            if '%'+key.strip()+'%' in p:
              print("Replacing %s with %s" % (key, interpolation_dict[key]))
              p = p.replace('%'+key.strip()+'%', interpolation_dict[key].strip())
        
        # Ok, have a path and host for the resource. Do we copy the file to or from?  
        if type == 'output':
          if h == '':
            # We want to copy this file from somewhere on the target host
            # to another location. 
            print("Warning, output copies of local file -> local file not implemented")
          elif h == 'localhost':
            # we want this file in the directory given on our local host.
            dir = os.path.dirname(p)
            if subprocess.call(["mkdir", "-p", dir]) != 0:
              self.logger.error('Could not copy file from host %s. Could not make local directory %s.' % (host, dir))
              continue
            else:
              self.logger.debug('Copying file. %s:%s to local file %s' %(host, file, p))
              with settings(**host_env), hide('output', 'running', 'warnings'):
                get(file, p)
          elif h != '':
            # Copy it on to the host.
            if subprocess.call(["mkdir", "-p", '/tmp/']) != 0:
              self.logger.error('Problem copying file from host %s to host %s. Could not copy to local /tmp' %(host,h))
            with settings(**host_env), hide('output', 'running', 'warnings'):
              get(file, '/tmp/'+p)
            dst_env = self.host_env(h)
            with settings(**dst_env), hide('output', 'running', 'warnings'):
              put('/tmp/'+p, p)
            # XXX Not removing local copy till fully tested.
            #subprocess.call(["rm", '/tmp/'+p])
            
        elif type == 'input':
          if h == '':
            with settings(**host_env), hide('output', 'running', 'warnings'):
              dir = os.path.dirname(file)
              if dir != '':
                result = run("mkdir -p dir")
                if result.failed:
                  result2 = sudo("mkdir -p dir")
                  if result2.failed:
                    self.logger.error('Failed to copy file %s to host %s. Could not make directory %s.' %(file, host, dir))
                    raise Exception('Failed to make directory on target host.')
                result3 = run("cp %s %s" % (p, file))
                if result3.failed:
                  self.logger.error('Failed to copy file %s to host %s. Did not copy.' %(file, host))
                              
          elif h == 'localhost':
            # look for this file locally, and copy it to the host:file
            try:
              with open(p) as f:
                f.close()
            except:
              self.logger.error('Problem finding file %s locally for host %s' % (p, host))
           
            with settings(**host_env), hide('output', 'running', 'warnings'):
              result = put(p, file)
              self.logger.debug('Copying file %s from local host to remote host %s:%s' % (p, host, file))
              if result.failure:
                self.logger.error('Failed to copy local file %s to %s:%s' % (p, host, file))
              
          elif h != '':
            # Find the file remotely and copy it to the host:file
            # XXX NOT IMPLEMENTED YET
            self.logger.error('Input from host to host is Not implemented yet.')
            continue
        
            
  def group_hosts_by_env(self, host_list):        
    """
    Returns a list of lists. Each list is a group of hosts which share login environments.
    """
    groups = []
    for host in host_list:
      placed = False
      try:
        host_env = self.host_env(host)
      except Exception as e:
        self.logger.error("Problem getting host environment for host \'%s\'." %(host))
        raise(e)
      for group in groups:    
        # Get a sample env of this group and compare to see if this current host fits in this group.
        try:
          group_env = self.host_env(group[0])
        except Exception as e:
          self.logger.error("Problem getting host for sample host \'%s\'." % (group[0]))
          raise(e)
        if(self.compare_envs(host_env, group_env)):	
          group.append(host)
          placed = True
          break
      if placed != True:
        # This host didn't fit with any other group. Make a new group for it.
        new_group = []
        new_group.append(host)
        groups.append(new_group)
    print groups
    return groups
  
  def compare_envs(self, env1, env2):  
    """
    Compares 2 envs and determines if they are essentially the same.
    """
    if len(env1) != len(env2):
      return False
    for key in env1:
      if key == 'host_string':
        # Host strings will be different. That's ok, the envs may still be the same.
        pass
      elif key not in env2:
        return False
      elif env1[key] != env2[key]:
        return False
    return True    


class OrchestratorFileHandler:
  def move_file(self, from_host_path, to_host_path):
    pass
    
  def file_exists(self, host_path):
    pass
  
  