import logging
import argparse
import OrchestratorHostsHandler
import OrchestratorGroupHandler
import OrchestratorBlockHandler
import OrchestratorConfgHandler

class Orchestrator:
  def __init__(self, args):
    self.args = args
    # Get our experiment configuration parser set up.
    try: 
      self.config_handler = OrchestratorConfgHandler.OrchestratorConfgHandler(args.config_file)
    except Exception, e:
      print(e)
      print('Exiting...')
      exit()
    
    
    # Get our logger going.
    try:
      if args.log_lvl != None:
        self.config_handler.set_logger(level=args.log_lvl)
      else:
        self.config_handler.set_logger()
    except Exception, e:
      print(e)
      print('Exiting...')
      exit()
    self.logger = logging.getLogger('Orchestrator')
    logger = logging.getLogger('Orchestrator')
    logger.info('')
    logger.info('********************** Beginning Experiment **********************')
    
    # Set up our host handler.
    self.host_handler = OrchestratorHostsHandler.OrchestratorHostsHandler(self.config_handler.return_hosts_config())
    # Test and collect passwords for hosts.
    print('Testing connections and collecting passwords for hosts.')
    self.host_handler.check_hosts()  
  
    # Set up our group handler.
    try:
      self.group_handler = OrchestratorGroupHandler.OrchestratorGroupHandler(self.config_handler.return_group_config())
    except Exception, e:
      print(e)
      print('Exiting...')
      exit()
    
    # Get a software block handler.
    if args.phase == 'extract':
      section = 'extraction'
    elif args.phase =='allocate':
      section ='resource allocation'
    elif args.phase == 'run':
      section ='experiment'
    else:
      raise Exception('Unknown experiment phase. Exiting...')
      exit()
    self.block_handler = OrchestratorBlockHandler.OrchestratorBlockHandler(self.config_handler.return_block_config(section), section)
    try:
      self.block_handler.load_all_blocks()
    except:
      raise
    
  def get_action_list(self):        
    # Were we given an action/actions on the command line?
    if self.args.action != None:
      self.logger.info('Given action items on command line (ignoring any actions in config): %s' % (self.args.action))    
      action_list = self.args.action
    # Else, do we have an action list in the config?
    else:
      action_list = self.block_handler.action_order()
    return action_list
  
  def get_block_list(self):
    try:
      blocks = self.block_handler.block_call_order()
    except Exception as e:
      raise e
    return blocks

  def get_target_hosts(self, block):
    """
      Gives the hosts a block should be run on.
      Returns None, [] or list of full host names.
      Returns an empty list if no tarets are specified for the block.
      It may be desirable to skip use of a block by commenting out targets.
    """
    # Get host targets for this block
    # Get the target (could be hosts or groups or both):
    targets = self.block_handler.get_target(block)
    target_hosts = []
    
    # Sort out the groups.
    for target in targets:
      # Is the target a group? Or a host?
      if self.group_handler.is_group(target):    
        # GROUP:
        # Get the full host name for all members of the group.
        for host in self.group_handler.host_members(target):
          # For each host, get the full name (and check that it's really a host we know about.)
          full_name = self.host_handler.full_name(host)
          if full_name != None:
            target_hosts.append(full_name)
          else:
            self.logger.error('Could not parse targets for block \'%s\'. Unknown target \'%s\'' % (block, host))
            raise Exception('Could not parse targets.')
      elif self.host_handler.is_host(self.host_handler.full_name(target)):  
        # HOST: This target wasn't a group, so we assume it's a host.
        full_name = self.host_handler.full_name(target)
        if target != None:
          target_hosts.append(full_name)
        else:
          self.logger.error('Could not parse targets for block \'%s\'. Unknown target \'%s\'' % (block, target))          
          raise Exception('Could not parse targets.')
      else:
        self.logger.error('Unknown target \'%s\'' % target)
        raise Exception('Could not parse targets.')

    # We may have some duplicates after all the group expanding, so remove dups.
    target_hosts = list(set(target_hosts))
    self.logger.info ("Block: %s -> Targets: %s"% (block, target_hosts))
    return target_hosts
    
  def do_action(self, block, action):
    # Get host targets for this block
    try:
      target_hosts = self.get_target_hosts(block)
    except:
      raise
    
    self.logger.info("Action \'%s\', Block \'%s\', Targets: \'%s\'"% (action, block, target_hosts))
        
    # Get function for this action
    try:
      function = self.block_handler.action_function(block, action)
    except:
      raise

    # Run action/function on targets!
    try:
      self.host_handler.run_on_host(target_hosts, function)
    except:
      raise
    
    
  
def main():
  parser = argparse.ArgumentParser(description='Software orchestrator to coordinate LegoTG')
  parser.add_argument("-p", "--phase", dest='phase', required=True, choices=['extract', 'allocate', 'run'])
  parser.add_argument("-a", "--action", dest='action', nargs="+", help='Action or list of actions to perform. Overrides default action list in experiment config.')
  parser.add_argument("-l", "--level" , dest='log_lvl', help='Set log level. Overrides log level set in experiment configuration file.')
  parser.add_argument("-c", "--check" , dest='checks', action="store_true")
  parser.add_argument("-f", "--config", dest='config_file', help="LegoTG experiment configuration file.")

  args = parser.parse_args()
  
  
  try:
    orchestrator = Orchestrator(args)
    logger = logging.getLogger('Orchestrator')
  except Exception, e:
    print(e)
    exit()
  try:
    block_list = orchestrator.get_block_list()
  except Exception as e:
    print(e)
    exit()
  if block_list != None:
    logger.info("List of blocks to call is: %s " % (block_list))
    try:
      action_list = orchestrator.get_action_list()    
    except Exception as e:
      print(e)
      exit()
    if action_list != None:
      for action in action_list:
        for block in block_list:
          try:
            orchestrator.do_action(block, action)  
          except Exception as e:
            print(e)
            exit()
  
if __name__ == "__main__":
  main()